unit Core;
{$WARN SYMBOL_PLATFORM OFF}
interface
uses Classes, Graphics;

const PixelsPerItem=20;
      MaxPlayers=6;
      LibraryFileName='N4.n4l';

type PFieldItem=^TFieldItem;
     TFieldItem=record
                  Owner:integer;
                  Fill:integer;
                  MaxFill:integer;
                  OldFill:integer;
                end;

type TCoord=record PX,PY:integer; end;

type TField=class
            private
              FSizeX,FSizeY,FSizeQ:integer;
              FPixelsWidth,FPixelsHeight:integer;
              Data:array of TFieldItem;
              procedure SetItem(X,Y:integer;Value:TFieldItem);
              function GetItem(X,Y:integer):TFieldItem;
              function GetItemPtr(X,Y:integer):PFieldItem;
            public
              Name:string;
              Canvas:TCanvas;
              Cursors:array[1..MaxPlayers]of TCoord;
              Count:array[-1..MaxPlayers]of integer;
              Wrap:boolean;
              HidePrefills:boolean;
              property Width:integer read FSizeX;
              property Height:integer read FSizeY;
              property Items:integer read FSizeQ;
              property PixelsWidth:integer read FPixelsWidth;
              property PixelsHeight:integer read FPixelsHeight;
              property Item[X,Y:integer]:TFieldItem read GetItem write SetItem; default;
              property ItemPtr[X,Y:integer]:PFieldItem read GetItemPtr;
              constructor Create; overload;
              constructor Create(_Width,_Height:integer;_Wrap,_Prefill:boolean); overload;
              constructor Create(_Name:string); overload;
              procedure InitField(_Wrap,_Prefill:boolean);
              function Increment(X,Y,Player:integer):boolean;
              function HandleExplosions:integer;
              procedure DrawItem(X,Y:integer);
              procedure DrawField;
              procedure SetCursor(Player,NewX,NewY:integer);
              procedure DoCount;
              function GetAIMove(Player:integer):TCoord;
              procedure WrapCoords(var X,Y:integer);
              procedure LoadFromLibrary;
              procedure SaveToLibrary;
              class function LibraryContents:TStringList;
            end;

implementation
uses SysUtils, SpriteMgr;

constructor TField.Create;
begin
  raise EAbstractError.Create('direct call to TField.Create');
end;

constructor TField.Create(_Width,_Height:integer;_Wrap,_Prefill:boolean);
var i:integer;
begin
  inherited Create;
  FSizeX:=_Width; FSizeY:=_Height;
  FSizeQ:=FSizeX*FSizeY;
  SetLength(Data,FSizeQ);
  for i:=0 to (FSizeQ-1) do
    with Data[i] do begin
      Owner:=0;
      Fill:=0;
      MaxFill:=0;
    end;
  HidePrefills:=false;
  InitField(_Wrap,_Prefill);
  FPixelsWidth:=FSizeX*PixelsPerItem;
  FPixelsHeight:=FSizeY*PixelsPerItem;
  Canvas:=nil;
  Name:='';
end;

constructor TField.Create(_Name:string);
begin
  inherited Create;
  FSizeX:=0; FSizeY:=0; FSizeQ:=0;
  Canvas:=nil;
  Name:=_Name;
  LoadFromLibrary;
end;

procedure TField.InitField(_Wrap,_Prefill:boolean);
var X,Y:integer;
begin
  Wrap:=_Wrap;
  for Y:=0 to (FSizeY-1) do
    for X:=0 to (FSizeX-1) do
      with Data[Y*FSizeX+X] do
        if Owner>=0 then begin
          MaxFill:=0;
          if GetItem(X-1,Y).Owner>=0 then inc(MaxFill);
          if GetItem(X+1,Y).Owner>=0 then inc(MaxFill);
          if GetItem(X,Y-1).Owner>=0 then inc(MaxFill);
          if GetItem(X,Y+1).Owner>=0 then inc(MaxFill);
          if MaxFill=0 then Owner:=-1 else begin
            Owner:=0;
            if _Prefill then Fill:=random(MaxFill)
                        else Fill:=0;
          end;
        end else
          MaxFill:=0;
  for X:=1 to 6 do
    with Cursors[X] do begin
      PX:=-1; PY:=-1;
    end;
end;

procedure TField.WrapCoords(var X,Y:integer);
begin
  if not Wrap then exit;
  if X<0 then inc(X,FSizeX);
  if Y<0 then inc(Y,FSizeY);
  if X>=FSizeX then dec(X,FSizeX);
  if Y>=FSizeY then dec(Y,FSizeY);
end;


procedure TField.SetItem(X,Y:integer;Value:TFieldItem);
begin
  if (X<0) OR (Y<0) OR (X>=FSizeX) OR (Y>=FSizeY) then exit;
  Data[Y*FSizeX+X]:=Value;
end;

function TField.GetItem(X,Y:integer):TFieldItem;
begin
  WrapCoords(X,Y);
  if (X<0) OR (Y<0) OR (X>=FSizeX) OR (Y>=FSizeY) then
    with Result do begin
      Owner:=-1;
      Fill:=0;
      MaxFill:=0;
    end
  else
    Result:=Data[Y*FSizeX+X];
end;

function TField.GetItemPtr(X,Y:integer):PFieldItem;
begin
  WrapCoords(X,Y);
  if (X<0) OR (Y<0) OR (X>=FSizeX) OR (Y>=FSizeY) then Result:=nil
                                                  else Result:=@Data[Y*FSizeX+X];
end;

procedure TField.DrawItem(X,Y:integer);
var BX,BY,P,F:integer; Item:TFieldItem;
begin
  if (X<0) OR (Y<0) OR (X>=FSizeX) OR (Y>=FSizeY) OR not(Assigned(Canvas)) then exit;
  Item:=Data[Y*FSizeX+X];
  BX:=X*PixelsPerItem;
  BY:=Y*PixelsPerItem;
  with Canvas,Item do begin
    if Owner<0 then begin
      Brush.Color:=$C0C0C0;
      FillRect(Rect(BX,BY,BX+PixelsPerItem,BY+PixelsPerItem))
    end else begin
      F:=Item.Fill;
      if F>10 then F:=10;
      if (Owner=0) AND HidePrefills then F:=0;
      Draw(BX,BY,Sprites[Owner,F]);
    end;
    for P:=1 to 6 do
      with Cursors[P] do
        if (PX=X) AND (PY=Y) then begin
          Draw(BX,BY,CursorSprites[P]);
          break;
        end;
  end;
end;

procedure TField.DrawField;
var X,Y:integer;
begin
  for Y:=0 to (FSizeY-1) do
    for X:=0 to (FSizeX-1) do
      DrawItem(X,Y);
end;

function TField.Increment(X,Y,Player:integer):boolean;
begin
  Result:=false;
  if (X<0) OR (X<0) OR (X>=FSizeX) OR (Y>=FSizeY) then exit;
  with Data[Y*FSizeX+X] do begin
    if (Owner<>0) AND (Owner<>Player) then exit;
    Owner:=Player;
    inc(Fill);
    DrawItem(X,Y);
  end;
  Result:=true;
end;

function TField.HandleExplosions:integer;
var ThePlayer:integer;
  procedure IncItem(X,Y:integer);
  var P:PFieldItem;
  begin
    WrapCoords(X,Y);
    P:=GetItemPtr(X,Y);
    if not Assigned(P) OR (P^.Owner<0) then exit;
    P^.Owner:=ThePlayer;
    inc(P^.Fill);
    DrawItem(X,Y);
  end;
var X,Y,I:integer;
begin
  Result:=0;
  for I:=0 to (FSizeQ-1) do
    with Data[I] do
      OldFill:=Fill;
  for Y:=0 to (FSizeY-1) do
    for X:=0 to (FSizeX-1) do
      with Data[Y*FSizeX+X] do
        if (Owner>=0) AND (OldFill>=MaxFill) then begin
          inc(Result);
          dec(Fill,MaxFill);
          DrawItem(X,Y);
          ThePlayer:=Owner;
          IncItem(X-1,Y);
          IncItem(X+1,Y);
          IncItem(X,Y-1);
          IncItem(X,Y+1);
        end;
end;

procedure TField.SetCursor(Player,NewX,NewY:integer);
var OldX,OldY:integer;
begin
  if (Player<1) OR (Player>MaxPlayers) then exit;
  with Cursors[Player] do begin
    OldX:=PX; OldY:=PY;
    PX:=NewX; PY:=NewY;
    DrawItem(OldX,OldY);
    DrawItem(NewX,NewY);
  end;
end;

procedure TField.DoCount;
var i:integer;
    P:PFieldItem;
begin
  FillChar(Count,sizeof(Count),0);
  P:=@Data[0];
  for i:=(FSizeQ-1) downto 0 do begin
    inc(Count[P^.Owner]);
    inc(P);
  end;
end;

function TField.GetAIMove(Player:integer):TCoord;
var X,Y:integer; h,Best,AlreadyFound:integer;
  function GetFill(X,Y:integer):integer;
  begin
    with GetItem(X,Y) do
      if (Owner=0) AND HidePrefills
        then Result:=0
        else Result:=Fill+(byte(Owner=0) AND 1);
  end;
begin
  Result.PX:=-1;
  Result.PY:=-1;
  Best:=-1;
  AlreadyFound:=0;
  for Y:=0 to (FSizeY-1) do
    for X:=0 to (FSizeX-1) do
      with Data[Y*FSizeX+X] do
        if (Owner=Player) OR (Owner=0) then begin
          h:=5*Fill+GetFill(X-1,Y)+GetFill(X+1,Y)
                   +GetFill(X,Y-1)+GetFill(X,Y+1);
          if h>Best then begin
            Best:=h;
            Result.PX:=X;
            Result.PY:=Y;
            AlreadyFound:=1;
          end else if h=Best then begin
            inc(AlreadyFound);
            if random(AlreadyFound)=1 then begin
              Result.PX:=X;
              Result.PY:=Y;
            end;
          end;
        end;
end;

procedure TField.LoadFromLibrary;
var T:system.text; s:string;
    Content:TStringList;
    Keep:boolean;
    i,X,Y:integer;
begin
  Content:=TStringList.Create;
  AssignFile(T,LibraryFileName);
  try
    Reset(T);
    Keep:=false;
    while not EOF(T) do begin
      readln(T,s);
      s:=Trim(s);
      if (length(s)>2) AND (s[1]='[') AND (s[length(s)]=']') then
        Keep:=(UpperCase(copy(s,2,length(s)-2))=UpperCase(Name));
      if Keep AND (length(s)>1) AND (s[1]<>'[') then begin
        Content.Add(s);
        if length(s)>FSizeX then FSizeX:=length(s);
      end;
    end;
    CloseFile(T);
  except
    // don't care if the file can't be read
  end;
  FSizeY:=Content.Count;
  FSizeQ:=FSizeX*FSizeY;
  SetLength(Data,FSizeQ);
  for i:=0 to (FSizeQ-1) do
    with Data[i] do begin
      Owner:=-1;
      Fill:=0;
      MaxFill:=0;
    end;
  for Y:=0 to (FSizeY-1) do begin
    s:=Content[Y];
    for X:=1 to length(s) do
      if not(s[X] IN [#32,'.','-']) then
        Data[Y*FSizeX+X-1].Owner:=0;
  end;
  HidePrefills:=false;
  InitField(false,false);
  FPixelsWidth:=FSizeX*PixelsPerItem;
  FPixelsHeight:=FSizeY*PixelsPerItem;
  Content.Free;
end;

procedure TField.SaveToLibrary;
var T:system.text; s:string;
    Keep:boolean;
    Content:TStringList;
    X,Y:integer;
begin
  Content:=TStringList.Create;
  AssignFile(T,LibraryFileName);
  try
    Reset(T);
    Keep:=true;
    while not EOF(T) do begin
      readln(T,s);
      s:=Trim(s);
      if (length(s)>2) AND (s[1]='[') AND (s[length(s)]=']') then
        Keep:=(UpperCase(copy(s,2,length(s)-2))<>UpperCase(Name));
      if Keep then Content.Add(s);
    end;
    CloseFile(T);
  except
    // don't care if the file can't be read
  end;
  Content.Add('['+Name+']');
  SetLength(s,FSizeX);
  for Y:=0 to (FSizeY-1) do begin
    for X:=0 to (FSizeX-1) do
      if Data[Y*FSizeX+X].Owner<0
        then s[X+1]:='.'
        else s[X+1]:='#';
    Content.Add(s);
  end;
  Content.Add('');
  try
    Content.SaveToFile(LibraryFileName);
  except
    // don't care, just like above (:
  end;
  Content.Free;
end;

class function TField.LibraryContents:TStringList;
var T:system.text; s:string;
begin
  Result:=TStringList.Create;
  AssignFile(T,LibraryFileName);
  try
    Reset(T);
    while not EOF(T) do begin
      readln(T,s);
      s:=Trim(s);
      if (length(s)>2) AND (s[1]='[') AND (s[length(s)]=']') then
        Result.Add(copy(s,2,length(s)-2));
    end;
    CloseFile(T);
  except
    // don't care if the file can't be read
  end;
end;

end.
