unit Main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, Core, Buttons;

const MaxExplosions=100000;

type TGameMode=(gmRound,gmAction,gmAsync);
     TPlayerType=(ptMouse,ptKey1,ptKey2,ptComp);

type TPlayerInfo=record
                   ID:integer;
                   PlayerType:TPlayerType;
                   Name:string;
                 end;

type
  TMainForm = class(TForm)
    TheBox: TPaintBox;
    ExplosionTimer: TTimer;
    BClose: TSpeedButton;
    BMinimize: TSpeedButton;
    BNew: TSpeedButton;
    AITimer: TTimer;
    PDeactivated: TPanel;
    BSave: TSpeedButton;
    BStart: TSpeedButton;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure TheBoxPaint(Sender: TObject);
    procedure TheBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ExplosionTimerTimer(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormPaint(Sender: TObject);
    procedure BCloseClick(Sender: TObject);
    procedure BMinimizeClick(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure AITimerTimer(Sender: TObject);
    procedure BNewClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure TheBoxMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure BStartClick(Sender: TObject);
    procedure BSaveClick(Sender: TObject);
  private
    { Private declarations }
    MoveSaveX,MoveSaveY:integer;
    LastX,LastY,SetTo:integer;
    procedure PostInit;
  public
    { Public declarations }
    Field:TField;
    MousePlayer,Key1Player,Key2Player:integer;
    Players:array[1..MaxPlayers]of TPlayerInfo;
    PlayerCount,CurrentPlayer:integer;
    GameMode:TGameMode;
    Running:boolean;
    EditMode:boolean;
    ExplosionInterval,ComputerMoveInterval:integer;
    procedure RedrawStatus(Full:boolean);
    procedure ProcessMove(Player,X,Y:integer);
  end;

var
  MainForm: TMainForm;

implementation
uses Math, SpriteMgr, NewGame;

{$R *.dfm}

procedure TMainForm.FormCreate(Sender: TObject);
var i:integer;
begin
  randomize;
  InitSprites;

  Field:=TField.Create(32,24,false,false);
  Field.HidePrefills:=false;
  EditMode:=false;
  GameMode:=gmAsync;
  ComputerMoveInterval:=1000;
  ExplosionInterval:=100;

  PlayerCount:=MaxPlayers;
  for i:=1 to PlayerCount do
    with Players[i] do begin
      ID:=i;
      PlayerType:=ptComp;
      Name:='Demospieler '+IntToStr(i);
    end;

  PostInit;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  Field.Free;
  FreeSprites;
end;

procedure TMainForm.PostInit;
var i,CompCount,MinWidth:integer;
begin
  Field.Canvas:=TheBox.Canvas;
  MinWidth:=Max(372,80*PlayerCount+144);
  TheBox.Width:=Field.PixelsWidth;
  TheBox.Height:=Field.PixelsHeight;
  ClientWidth:=Max(MinWidth,Field.PixelsWidth+4);
  ClientHeight:=Field.PixelsHeight+50;
  TheBox.Left:=(ClientWidth-Field.PixelsWidth) DIV 2;

  MousePlayer:=0; Key1Player:=0; Key2Player:=0;

  BStart.Visible:=EditMode;
  if EditMode then begin
    RedrawStatus(True);
    PDeactivated.Visible:=false;
    ExplosionTimer.Enabled:=false;
    AITimer.Enabled:=false;
    TheBox.Invalidate;
    exit;
  end;

  if GameMode<>gmRound then begin
    for i:=1 to PlayerCount do
      case Players[i].PlayerType of
        ptMouse: MousePlayer:=Players[i].ID;
        ptKey1: Key1Player:=Players[i].ID;
        ptKey2: Key2Player:=Players[i].ID;
      end;
    if Key1Player<>0 then
      with Field.Cursors[Key1Player] do begin
        PX:=random(Field.Width);
        PY:=random(Field.Height);
      end;
    if Key2Player<>0 then
      with Field.Cursors[Key2Player] do begin
        PX:=random(Field.Width);
        PY:=random(Field.Height);
      end;
  end else
    CurrentPlayer:=1;

  with AITimer do
    if GameMode=gmRound then begin
      Interval:=ComputerMoveInterval;
      Tag:=1;
      Enabled:=(Players[1].PlayerType=ptComp);
    end else begin
      CompCount:=0;
      for i:=1 to PlayerCount do
        if Players[i].PlayerType=ptComp then
          inc(CompCount);
      if CompCount=0 then
        Enabled:=false
      else begin
        Enabled:=true;
        Interval:=ComputerMoveInterval DIV CompCount;
        Tag:=1;
      end;
    end;

  Running:=true;
  RedrawStatus(True);
  PDeactivated.Visible:=false;
  ExplosionTimer.Interval:=ExplosionInterval;
  ExplosionTimer.Enabled:=(GameMode=gmAsync);
  TheBox.Invalidate;
end;

procedure TMainForm.TheBoxPaint(Sender: TObject);
begin
  Field.DrawField;
end;

procedure TMainForm.ProcessMove(Player,X,Y:integer);
var ExplNow,ExplTotal:integer;
    OK:boolean;
begin
  if not(Running) OR (Player<1) OR (Player>MaxPlayers) then exit;
  OK:=Field.Increment(X,Y,Player);
  // synchrone Explosionen durchfhren
  if GameMode<>gmAsync then begin
    ExplTotal:=0;
    repeat
      ExplNow:=Field.HandleExplosions;
      inc(ExplTotal,ExplNow);
    until (ExplNow=0) OR (ExplTotal>MaxExplosions);
    if ExplTotal>MaxExplosions then Running:=false;
  end;
  // Spiel beendet -> Meldung anzeigen
  if not(Running) then begin
    PDeactivated.Left:=(ClientWidth-PDeactivated.Width) DIV 2;
    PDeactivated.Top:=48+(Field.PixelsHeight-PDeactivated.Height) DIV 2;
    PDeactivated.Visible:=true;
    AITimer.Enabled:=false;
  end;
  // nchster Spieler + Status-Redraw
  if Running AND OK AND (GameMode=gmRound) then begin
    CurrentPlayer:=(CurrentPlayer MOD PlayerCount)+1;
    if Players[CurrentPlayer].PlayerType=ptComp then
      with AITimer do begin
        Tag:=CurrentPlayer;
        Enabled:=true;
      end;
  end;
  RedrawStatus(false);
end;

procedure TMainForm.TheBoxMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  X:=X DIV PixelsPerItem;
  Y:=Y DIV PixelsPerItem;
  if EditMode then begin
    LastX:=X; LastY:=Y;
    with Field.ItemPtr[X,Y]^ do begin
      SetTo:=Owner XOR (-1);
      Owner:=SetTo;
    end;
    Field.DrawItem(X,Y);
  end else if GameMode<>gmRound then
    ProcessMove(MousePlayer,X,Y)
  else if Players[CurrentPlayer].PlayerType<>ptComp then
    ProcessMove(Players[CurrentPlayer].ID,X,Y);
end;

procedure TMainForm.TheBoxMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var I:PFieldItem;
begin
  if not(EditMode) OR not(ssLeft IN Shift) then exit;
  X:=X DIV PixelsPerItem;
  Y:=Y DIV PixelsPerItem;
  if (X=LastX) AND (Y=LastY) then exit;
  LastX:=X; LastY:=Y;
  with Field do begin
    I:=ItemPtr[X,Y];
    if Assigned(I) then
      I^.Owner:=SetTo;
    DrawItem(X,Y);
  end;
end;

procedure TMainForm.ExplosionTimerTimer(Sender: TObject);
begin
  Field.HandleExplosions;
  RedrawStatus(false);
end;

procedure TMainForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
  procedure CMove(Player,RX,RY:integer);
  var NX,NY:integer;
  begin
    if Player<1 then exit;
    with Field.Cursors[Player] do begin
      NX:=PX+RX; NY:=PY+RY;
    end;
    Field.WrapCoords(NX,NY);
    with Field do
      if (NX>=0) AND (NY>=0) AND (NX<Width) AND (NY<Height) then
        SetCursor(Player,NX,NY);
  end;
begin
  case Key of
    VK_UP:    CMove(Key1Player, 0,-1);
    VK_DOWN:  CMove(Key1Player, 0,+1);
    VK_LEFT:  CMove(Key1Player,-1, 0);
    VK_RIGHT: CMove(Key1Player,+1, 0);
    13:       if Key1Player>0 then with Field.Cursors[Key1Player] do
              ProcessMove(Key1Player,PX,PY);
    Ord('W'): CMove(Key2Player, 0,-1);
    Ord('S'): CMove(Key2Player, 0,+1);
    Ord('A'): CMove(Key2Player,-1, 0);
    Ord('D'): CMove(Key2Player,+1, 0);
    Ord('Q'): if Key2Player>0 then with Field.Cursors[Key2Player] do
              ProcessMove(Key2Player,PX,PY);
  end;
end;

procedure TMainForm.FormPaint(Sender: TObject);
var SX,SY:integer;
begin with Canvas do begin
  SX:=ClientWidth;
  SY:=ClientHeight;
  Draw(4,4,N4Logo);
  Pen.Style:=psSolid;
  Pen.Width:=1;
  Pen.Color:=$FFFFFF; MoveTo(0,SY-1); LineTo(0,0); LineTo(SX,0);
  Pen.Color:=$404040; MoveTo(SX-1,1); LineTo(SX-1,SY-1); LineTo(0,SY-1);
  Pen.Color:=$E0E0E0; MoveTo(1,SY-2); LineTo(1,1); LineTo(SX-1,1);
  Pen.Color:=$808080; MoveTo(SX-2,2); LineTo(SX-2,SY-2); LineTo(1,SY-2);
  RedrawStatus(true);
end; end;

procedure TMainForm.BCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.BMinimizeClick(Sender: TObject);
begin
  Application.Minimize;
end;

procedure TMainForm.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if Button<>mbLeft then exit;
  MoveSaveX:=X;
  MoveSaveY:=Y;
end;

procedure TMainForm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if not (ssLeft IN Shift) then exit;
  Left:=Left+X-MoveSaveX;
  Top:=Top+Y-MoveSaveY;
end;

procedure TMainForm.RedrawStatus(Full:boolean);
var i,w2d:integer;
begin with Canvas do begin
  if Full then FillRect(Rect(42,25,ClientWidth-2,48));
  Field.DoCount;
  // Anzahl belegter Felder
  Font.Style:=[];
  for i:=1 to PlayerCount do
    TextOut(i*100-36,36,IntToStr(Field.Count[Players[i].ID])+'      ');
  // Spielernamen
  if Full then begin
    Font.Style:=[fsBold];
    for i:=1 to PlayerCount do
      TextOut(i*100-36,25,Players[i].Name);
  end;
  // Spielerfarben
  if Full OR (GameMode=gmRound) then
    for i:=1 to PlayerCount do begin
      if not(EditMode) AND ((GameMode<>gmRound) OR not(Active) OR (i=CurrentPlayer))
        then w2d:=Players[i].ID else w2d:=0;
      Draw(i*100-58,26,Sprites[w2d,0]);
    end;
end; end;

procedure TMainForm.AITimerTimer(Sender: TObject);
var P:integer;
    Pos:TCoord;
begin
  if GameMode=gmRound then AITimer.Enabled:=false;
  P:=AITimer.Tag;
  while Players[P].PlayerType<>ptComp do
    P:=(P MOD PlayerCount)+1;
  with Players[P] do begin
    Pos:=Field.GetAIMove(ID);
    ProcessMove(ID,Pos.PX,Pos.PY);
  end;
  AITimer.Tag:=(P MOD PlayerCount)+1;
end;

procedure TMainForm.BNewClick(Sender: TObject);
begin
  if NewGameForm.ShowModal=mrOK then PostInit;
end;

procedure TMainForm.FormResize(Sender: TObject);
begin
  Invalidate;
end;

procedure TMainForm.BStartClick(Sender: TObject);
begin
  EditMode:=not(EditMode);
  with Field do InitField(Wrap,NewGameForm.CPrefill.Checked);
  PostInit;
end;

procedure TMainForm.BSaveClick(Sender: TObject);
var NewName:string;
begin
  NewName:=Field.Name;
  if not InputQuery('Spielfeld speichern',
                    'Unter welchem Namen soll das Spielfeld abgelegt werden?',
                    NewName) then exit;
  NewName:=Trim(NewName);
  if NewName='' then exit;
  Field.Name:=NewName;
  Field.SaveToLibrary;
end;

end.
