unit NewGame;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, ExtCtrls, Buttons, Core, Main;

type
  TNewGameForm = class(TForm)
    CMode: TComboBox;
    LMode: TLabel;
    LField: TLabel;
    FDefault: TRadioButton;
    FLoad: TRadioButton;
    CField: TComboBox;
    ESizeX: TEdit;
    CSizeX: TUpDown;
    ESizeY: TEdit;
    CSizeY: TUpDown;
    Lx: TLabel;
    CWrap: TCheckBox;
    CPrefill: TCheckBox;
    CHidePrefills: TCheckBox;
    CEdit: TCheckBox;
    GPlayers: TGroupBox;
    PPlayerCount: TPanel;
    Panel1: TPanel;
    Panel2: TPanel;
    P1P: TPanel;
    P1C: TPaintBox;
    P1N: TEdit;
    P1pmMouse: TSpeedButton;
    P1pmComp: TSpeedButton;
    P1pmKey1: TSpeedButton;
    P1pmKey2: TSpeedButton;
    P2P: TPanel;
    P2C: TPaintBox;
    P2pmMouse: TSpeedButton;
    P2pmComp: TSpeedButton;
    P2pmKey1: TSpeedButton;
    P2pmKey2: TSpeedButton;
    P2N: TEdit;
    P3P: TPanel;
    P3C: TPaintBox;
    P3pmMouse: TSpeedButton;
    P3pmComp: TSpeedButton;
    P3pmKey1: TSpeedButton;
    P3pmKey2: TSpeedButton;
    P3N: TEdit;
    P4P: TPanel;
    P4C: TPaintBox;
    P4pmMouse: TSpeedButton;
    P4pmComp: TSpeedButton;
    P4pmKey1: TSpeedButton;
    P4pmKey2: TSpeedButton;
    P4N: TEdit;
    P5P: TPanel;
    P5C: TPaintBox;
    P5pmMouse: TSpeedButton;
    P5pmComp: TSpeedButton;
    P5pmKey1: TSpeedButton;
    P5pmKey2: TSpeedButton;
    P5N: TEdit;
    P6P: TPanel;
    P6C: TPaintBox;
    P6pmMouse: TSpeedButton;
    P6pmComp: TSpeedButton;
    P6pmKey1: TSpeedButton;
    P6pmKey2: TSpeedButton;
    P6N: TEdit;
    PColorSelect: TPanel;
    BColorSelect: TPaintBox;
    bvColorSelect: TBevel;
    GTiming: TGroupBox;
    LExplosionInterval: TLabel;
    CExplosionInterval: TTrackBar;
    LComputerMoveInterval: TLabel;
    CComputerMoveInterval: TTrackBar;
    BOK: TPanel;
    BCancel: TPanel;
    procedure FormCreate(Sender: TObject);
    procedure PxCPaint(Sender: TObject);
    procedure P1CMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure BColorSelectPaint(Sender: TObject);
    procedure BColorSelectMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PxpmCompClick(Sender: TObject);
    procedure PxpmMouseClick(Sender: TObject);
    procedure PxpmKey1Click(Sender: TObject);
    procedure PxpmKey2Click(Sender: TObject);
    procedure PlayerCountChange(Sender: TObject);
    procedure VirtualButtonMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure VirtualButtonMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SomethingChanged(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure CExplosionIntervalChange(Sender: TObject);
    procedure CComputerMoveIntervalChange(Sender: TObject);
    procedure BCancelClick(Sender: TObject);
    procedure BOKClick(Sender: TObject);
  private
    { Private declarations }
    GameMode:TGameMode;
    PlayerCount:integer;
    ExplosionInterval,ComputerMoveInterval:integer;
    ColorAssign:array[1..MaxPlayers]of integer;
    PlayerTypes:array[1..MaxPlayers]of TPlayerType;
    procedure Validate;
  public
    { Public declarations }
  end;

var
  NewGameForm: TNewGameForm;

implementation
uses SpriteMgr;

{$R *.dfm}

procedure TNewGameForm.FormCreate(Sender: TObject);
var i:integer;
begin
  PlayerCount:=6;
  for i:=1 to MaxPlayers do ColorAssign[i]:=i;
  for i:=1 to MaxPlayers do PlayerTypes[i]:=ptComp;
  CExplosionIntervalChange(nil);
  CComputerMoveIntervalChange(nil);
end;

procedure TNewGameForm.PxCPaint(Sender: TObject);
begin
  with Sender as TPaintBox,Canvas do
    Draw(0,0,Sprites[ColorAssign[Tag],Tag]);
end;

procedure TNewGameForm.P1CMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if Button<>mbLeft then exit;
  with Sender as TPaintBox do begin
    PColorSelect.Top:=Self.ScreenToClient(ClientToScreen(Point(0,0))).Y-2;
    PColorSelect.Visible:=true;
    BColorSelect.Tag:=Tag;
  end;
end;

procedure TNewGameForm.BColorSelectPaint(Sender: TObject);
var i:integer;
begin
  with BColorSelect.Canvas do
    for i:=0 to (MaxPlayers-1) do
      Draw(i*PixelsPerItem,0,Sprites[i+1,0]);
end;

procedure TNewGameForm.BColorSelectMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var Slot,Other,OldColor,NewColor:integer;
begin
  if Button<>mbLeft then exit;
  Slot:=BColorSelect.Tag;
  OldColor:=ColorAssign[Slot];
  NewColor:=(X DIV PixelsPerItem)+1;
  for Other:=1 to MaxPlayers do
    if ColorAssign[Other]=NewColor then break;
  ColorAssign[Slot]:=NewColor;
  ColorAssign[Other]:=OldColor;
  PColorSelect.Visible:=false;
  P1C.Invalidate; P2C.Invalidate; P3C.Invalidate;
  P4C.Invalidate; P5C.Invalidate; P6C.Invalidate;
end;

procedure __set(mo:TPlayerType;c,m,k1,k2:TSpeedButton);
begin case mo of
  ptComp: c.Down:=true;
  ptMouse:m.Down:=true;
  ptKey1:k1.Down:=true;
  ptKey2:k2.Down:=true;
end; end;

procedure TNewGameForm.Validate;
var i:integer; HaveMouse,HaveKey1,HaveKey2,z:boolean;
begin
  // Spielmodus?
  case CMode.ItemIndex of
    0:GameMode:=gmRound;
    1:GameMode:=gmAction;
    2:GameMode:=gmAsync;
  end;
  // Kosmetik
  z:=FDefault.Checked;
  ESizeX.Enabled:=z; CSizeX.Enabled:=z; Lx.Enabled:=z;
  ESizeY.Enabled:=z; CSizeY.Enabled:=z;
  CField.Enabled:=not(z);
  CHidePrefills.Enabled:=CPrefill.Checked;
  LExplosionInterval.Enabled:=(GameMode=gmAsync);
  CExplosionInterval.Enabled:=(GameMode=gmAsync);
  LComputerMoveInterval.Enabled:=(GameMode<>gmRound);
  CComputerMoveInterval.Enabled:=(GameMode<>gmRound);
  // PlayerType-Constraints
  HaveMouse:=false; HaveKey1:=false; HaveKey2:=false;
  for i:=1 to MaxPlayers do
    if GameMode=gmRound then begin
      if PlayerTypes[i] IN [ptKey1,ptKey2] then PlayerTypes[i]:=ptMouse;
    end else case PlayerTypes[i] of
      ptMouse: if HaveMouse then PlayerTypes[i]:=ptComp else HaveMouse:=true;
      ptKey1:  if HaveKey1  then PlayerTypes[i]:=ptComp else HaveKey1:=true;
      ptKey2:  if HaveKey2  then PlayerTypes[i]:=ptComp else HaveKey2:=true;
    end;
  // PlayerTypes bernehmen
  __set(PlayerTypes[1],P1pmComp,P1pmMouse,P1pmKey1,P1pmKey2);
  __set(PlayerTypes[2],P2pmComp,P2pmMouse,P2pmKey1,P2pmKey2);
  __set(PlayerTypes[3],P3pmComp,P3pmMouse,P3pmKey1,P3pmKey2);
  __set(PlayerTypes[4],P4pmComp,P4pmMouse,P4pmKey1,P4pmKey2);
  __set(PlayerTypes[5],P5pmComp,P5pmMouse,P5pmKey1,P5pmKey2);
  __set(PlayerTypes[6],P6pmComp,P6pmMouse,P6pmKey1,P6pmKey2);
end;

procedure TNewGameForm.PxpmCompClick(Sender: TObject);
begin
  PlayerTypes[(Sender as TControl).Tag]:=ptComp;
  Validate;
end;

procedure TNewGameForm.PxpmMouseClick(Sender: TObject);
begin
  PlayerTypes[(Sender as TControl).Tag]:=ptMouse;
  Validate;
end;

procedure TNewGameForm.PxpmKey1Click(Sender: TObject);
begin
  PlayerTypes[(Sender as TControl).Tag]:=ptKey1;
  Validate;
end;

procedure TNewGameForm.PxpmKey2Click(Sender: TObject);
begin
  PlayerTypes[(Sender as TControl).Tag]:=ptKey2;
  Validate;
end;

procedure TNewGameForm.PlayerCountChange(Sender: TObject);
begin
  PlayerCount:=PlayerCount+(Sender as TComponent).Tag;
  if PlayerCount<1 then PlayerCount:=1;
  if PlayerCount>6 then PlayerCount:=6;
  PPlayerCount.Caption:=IntToStr(PlayerCount);
  P2P.Visible:=(PlayerCount>=2);
  P3P.Visible:=(PlayerCount>=3);
  P4P.Visible:=(PlayerCount>=4);
  P5P.Visible:=(PlayerCount>=5);
  P6P.Visible:=(PlayerCount>=6);
end;

procedure TNewGameForm.VirtualButtonMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  (Sender as TPanel).BevelOuter:=bvLowered;
end;

procedure TNewGameForm.VirtualButtonMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  (Sender as TPanel).BevelOuter:=bvRaised;
end;

procedure TNewGameForm.SomethingChanged(Sender: TObject);
begin
  Validate;
end;

procedure TNewGameForm.FormShow(Sender: TObject);
var i:integer;
begin
  CField.Clear;
  CField.Items.Add('[- aktuelles Spielfeld -]');
  CField.ItemIndex:=0;
  with TField.LibraryContents do begin
    CaseSensitive:=false;
    Sort;
    for i:=0 to (Count-1) do
      CField.Items.Add(Strings[i]);
    Free;
  end;
  Validate;
end;

procedure TNewGameForm.CExplosionIntervalChange(Sender: TObject);
var V,i:integer;
begin
  V:=CExplosionInterval.Position;
  case V MOD 6 of
    0: ExplosionInterval:=10;
    1: ExplosionInterval:=20;
    2: ExplosionInterval:=35;
    3: ExplosionInterval:=33;
    4: ExplosionInterval:=50;
    5: ExplosionInterval:=75;
  end;
  for i:=(V DIV 6) downto 1 do
    ExplosionInterval:=ExplosionInterval*10;
  LExplosionInterval.Caption:=
    'Explosionen aller '+IntToStr(ExplosionInterval)+' ms';
end;

procedure TNewGameForm.CComputerMoveIntervalChange(Sender: TObject);
var V,i:integer;
begin
  V:=CComputerMoveInterval.Position;
  case V MOD 6 of
    0: ComputerMoveInterval:=50;
    1: ComputerMoveInterval:=75;
    2: ComputerMoveInterval:=100;
    3: ComputerMoveInterval:=200;
    4: ComputerMoveInterval:=250;
    5: ComputerMoveInterval:=330;
  end;
  for i:=(V DIV 6) downto 1 do
    ComputerMoveInterval:=ComputerMoveInterval*10;
  LComputerMoveInterval.Caption:=
    'Computerzge aller '+IntToStr(ComputerMoveInterval)+' ms';
end;

procedure TNewGameForm.BCancelClick(Sender: TObject);
begin
  ModalResult:=mrCancel;
end;

procedure TNewGameForm.BOKClick(Sender: TObject);
var i:integer;
begin
  with MainForm do begin
    EditMode:=CEdit.Checked;
    if FDefault.Checked then begin
      Field.Free;
      Field:=TField.Create(CSizeX.Position,CSizeY.Position,CWrap.Checked,CPrefill.Checked AND not(EditMode));
    end else if CField.ItemIndex=0 then
      Field.InitField(CWrap.Checked,CPrefill.Checked AND not(EditMode))
    else begin
      Field.Free;
      Field:=TField.Create(CField.Items[CField.ItemIndex]);
      Field.InitField(CWrap.Checked,CPrefill.Checked AND not(EditMode))
    end;
    // Verschiedenes
    Field.HidePrefills:=CHidePrefills.Checked;
    GameMode:=Self.GameMode;
    ExplosionInterval:=Self.ExplosionInterval;
    ComputerMoveInterval:=Self.ComputerMoveInterval;
    BSave.Visible:=EditMode;
    // Spielereinstellungen
    PlayerCount:=Self.PlayerCount;
    for i:=1 to PlayerCount do
      with Players[i] do begin
        ID:=ColorAssign[i];
        PlayerType:=PlayerTypes[i];
      end;
    // Spielernamen
    Players[1].Name:=P1N.Text;
    Players[2].Name:=P2N.Text;
    Players[3].Name:=P3N.Text;
    Players[4].Name:=P4N.Text;
    Players[5].Name:=P5N.Text;
    Players[6].Name:=P6N.Text;
  end;
  ModalResult:=mrOK;
end;

end.
