unit Saver;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls;

type PCell=^TCell;
     TCell=record
             Owner,LastOwner:integer;
             Fill,LastFill:integer;
             MaxFill:integer;
           end;

type
  TSaverForm = class(TForm)
    TickTimer: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure TickTimerTimer(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
  private
    { Private declarations }
    MouseRef:TPoint;
    ScreenX,ScreenY:integer;
    OffsetX,OffsetY:integer;
    CellSize:integer;
    WrapX,WrapY:boolean;
    CellsX,CellsY,CellsQ:integer;
    Players,CurrentPlayer:integer;
    MoveEvery,Countdown:integer;
    StableTime:integer;
    Deconstruct:boolean;
    DeconPos,DeconIncr,DeconSpeed,DeconToDo:integer;
    Sprites:array of TBitmap;
    Field:array of TCell;
    function Cell(X,Y:integer):PCell;
    procedure Init;
    procedure InitDeconstruction;
  public
    { Public declarations }
  end;

var
  SaverForm: TSaverForm;

implementation
uses Math;

{$R *.dfm}

procedure TSaverForm.FormCreate(Sender: TObject);
begin
//AllocConsole;
  randomize;
  ScreenX:=Screen.Width;
  ScreenY:=Screen.Height;
  Top:=0; Left:=0;
  Width:=ScreenX; Height:=ScreenY;
  GetCursorPos(MouseRef);
  SetLength(Sprites,0);
  Init;
end;

procedure TSaverForm.FormDestroy(Sender: TObject);
begin
  BorderStyle:=bsSizeable;
  Width:=10; Height:=10;
end;

procedure TSaverForm.Init;
var i,j,R,G,B,X,Y,Z,Half:integer;
    L:PByteArray; P:^byte;
    K:array of integer;
    f,yp,hstart,pwr:real;
    HaveColors:array[0..5]of boolean;
  function echo(i:integer):integer;
  begin if (i>=Half) then Result:=i-CellSize+Half else Result:=Half-i-1; end;
begin
  CellSize:=6+random(34);
  CellsX:=ScreenX DIV CellSize;
  CellsY:=ScreenY DIV CellSize;
  CellsQ:=CellsX*CellsY;
  SetLength(Field,CellsQ);
  WrapX:=(random(2)<>0);
  WrapY:=(random(2)<>0);
  for Y:=0 to (CellsY-1) do
    for X:=0 to (CellsX-1) do
      with Field[Y*CellsX+X] do begin
        Owner:=-1;
        LastOwner:=-1;
        MaxFill:=0;
        if Cell(X-1,Y)<>nil then inc(MaxFill);
        if Cell(X+1,Y)<>nil then inc(MaxFill);
        if Cell(X,Y-1)<>nil then inc(MaxFill);
        if Cell(X,Y+1)<>nil then inc(MaxFill);
        Fill:=random(MaxFill);
      end;
  OffsetX:=(ScreenX-CellsX*CellSize) DIV 2;
  OffsetY:=(ScreenY-CellsY*CellSize) DIV 2;

  Players:=random(5)+2;
  SetLength(Sprites,Players);
  R:=0; G:=0; B:=0;
  Half:=(CellSize+1) SHR 1;
  SetLength(K,Half*Half);
  hstart:=1.0-2.0/(CellSize-1);
  hstart:=hstart-sqr(random)*hstart;
  for i:=0 to 5 do HaveColors[i]:=false;
  pwr:=2.0+sqr(sqr(random))*5.0;

  Z:=0;
  for Y:=0 to (Half-1) do begin
    yp:=Power((Y+0.5)/(Half-0.5),pwr);
    for X:=0 to (Half-1) do begin
      f:=Power(yp+Power((X+0.5)/(Half-0.5),pwr),1.0/pwr);
      f:=(f-hstart)/(1.0-hstart);
      if f<0.0 then f:=0.0;
      if f>1.0 then K[Z]:=0
               else K[Z]:=round(257.0*(1.0-f));
      inc(Z);
    end;
  end;

  for i:=0 to (Players-1) do begin
    X:=random(128);
    Y:=random(256);
    Z:=random(64)+192;
    repeat j:=random(6); until not HaveColors[j];
    HaveColors[j]:=true;
    case j of
      0:begin R:=X; G:=Y; B:=Z; end;
      1:begin R:=X; G:=Z; B:=Y; end;
      2:begin R:=Y; G:=X; B:=Z; end;
      3:begin R:=Y; G:=Z; B:=X; end;
      4:begin R:=Z; G:=X; B:=Y; end;
      5:begin R:=Z; G:=Y; B:=X; end;
    end;
    Sprites[i]:=TBitmap.Create;
    with Sprites[i] do begin
      Width:=CellSize; Height:=CellSize;
      PixelFormat:=pf24bit;
      for Y:=0 to (CellSize-1) do begin
        L:=Sprites[i].ScanLine[Y]; P:=@L[0];
        Z:=echo(Y)*Half;
        for X:=0 to (CellSize-1) do begin
          j:=Z+echo(X);
          P^:=(B*K[j]) SHR 8; inc(P);
          P^:=(G*K[j]) SHR 8; inc(P);
          P^:=(R*K[j]) SHR 8; inc(P);
        end;
      end;
      PixelFormat:=pfDevice;
    end;
  end;

  TickTimer.Interval:=15+random(65);
  TickTimer.Enabled:=true;
  MoveEvery:=((50+random(450)) DIV Players DIV integer(TickTimer.Interval))+1;
  Countdown:=1;
  CurrentPlayer:=0;
  StableTime:=0;
  Deconstruct:=false;
end;

function TSaverForm.Cell(X,Y:integer):PCell;
begin
  if WrapX then begin
    while X<0 do inc(X,CellsX);
    while X>=CellsX do dec(X,CellsX);
  end;
  if WrapY then begin
    while Y<0 do inc(Y,CellsY);
    while Y>=CellsY do dec(Y,CellsY);
  end;
  if (Y>=0) AND (X>=0) AND (X<CellsX) AND (Y<CellsY)
    then Result:=@Field[Y*CellsX+X]
    else Result:=nil;
end;

procedure TSaverForm.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  Close;
end;

procedure TSaverForm.TickTimerTimer(Sender: TObject);
var i,X,Y,H,me:integer;
    C:PCell;
    BestX,BestY,BestH:integer;
    OfThisClass:integer;
const IWasHere=-123;
begin
  if Deconstruct then begin
    Canvas.Brush.Style:=bsSolid;
    Canvas.Brush.Color:=0;
    for me:=DeconSpeed downto 0 do begin
      while Field[DeconPos].Owner=IWasHere do DeconPos:=(DeconPos+1) MOD CellsQ;
      Field[DeconPos].Owner:=IWasHere;
      X:=OffsetX+CellSize*(DeconPos MOD CellsX);
      Y:=OffsetY+CellSize*(DeconPos DIV CellsX);
      Canvas.FillRect(Rect(X,Y,X+CellSize,Y+CellSize));
      dec(DeconToDo);
      if DeconToDo=0 then begin
        Init;
        exit;
      end else
        DeconPos:=(DeconPos+DeconIncr) MOD CellsQ;
    end;
    exit;
  end;

  dec(Countdown);
  if Countdown<1 then begin
    Countdown:=MoveEvery;
    CurrentPlayer:=(CurrentPlayer+1) MOD Players;
    BestX:=0; BestY:=0; BestH:=-1; OfThisClass:=0;
    for Y:=0 to (CellsY-1) do
      for X:=0 to (CellsX-1) do
        with Field[Y*CellsX+X] do
          if (Owner<0) OR (Owner=CurrentPlayer) then begin
            H:=4*Fill;
            C:=Cell(X-1,Y); if C<>nil then with C^ do begin inc(H,Fill); if Owner<0 then inc(H); end;
            C:=Cell(X+1,Y); if C<>nil then with C^ do begin inc(H,Fill); if Owner<0 then inc(H); end;
            C:=Cell(X,Y-1); if C<>nil then with C^ do begin inc(H,Fill); if Owner<0 then inc(H); end;
            C:=Cell(X,Y+1); if C<>nil then with C^ do begin inc(H,Fill); if Owner<0 then inc(H); end;
            if H>BestH then begin
              BestX:=X; BestY:=Y; BestH:=H;
              OfThisClass:=1;
            end else if H=BestH then begin
              inc(OfThisClass);
              if random(OfThisClass)=1 then begin
                BestX:=X; BestY:=Y;
              end;
            end;
          end;
    if BestH>=0 then
      with Field[BestY*CellsX+BestX] do begin
        Owner:=CurrentPlayer;
        inc(Fill);
      end;
  end;

  for i:=0 to (CellsQ-1) do
    with Field[i] do
      LastFill:=Fill;
  for Y:=0 to (CellsY-1) do
    for X:=0 to (CellsX-1) do
      with Field[Y*CellsX+X] do
        if (Owner>=0) AND (LastFill>=MaxFill) then begin
          dec(Fill,MaxFill);
          me:=Owner;
          C:=Cell(X-1,Y); if C<>nil then with C^ do begin inc(Fill); Owner:=me; end;
          C:=Cell(X+1,Y); if C<>nil then with C^ do begin inc(Fill); Owner:=me; end;
          C:=Cell(X,Y-1); if C<>nil then with C^ do begin inc(Fill); Owner:=me; end;
          C:=Cell(X,Y+1); if C<>nil then with C^ do begin inc(Fill); Owner:=me; end;
        end;
  H:=0;
  for Y:=0 to (CellsY-1) do
    for X:=0 to (CellsX-1) do
      with Field[Y*CellsX+X] do
        if Owner<>LastOwner then begin
          Canvas.Draw(OffsetX+X*CellSize,OffsetY+Y*CellSize,Sprites[Owner]);
          LastOwner:=Owner;
          inc(H);
        end;
  if H>0 then StableTime:=0
         else inc(StableTime,TickTimer.Interval);
  if StableTime>1500 then InitDeconstruction;
end;

procedure TSaverForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key=VK_PAUSE then InitDeconstruction
  else Close;
end;

procedure TSaverForm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var Mouse:TPoint;
begin
  GetCursorPos(Mouse);
  if (abs(Mouse.X-MouseRef.X)+abs(Mouse.Y-MouseRef.Y))>5 then Close;
end;

procedure TSaverForm.InitDeconstruction;
begin
  Deconstruct:=true;
  DeconPos:=random(CellsQ);
  DeconIncr:=random(CellsQ);
  TickTimer.Interval:=40;
  DeconSpeed:=(CellsQ DIV (25+random(50)))+1;
  DeconToDo:=CellsQ;
end;

end.
